@@ -1,7 +1,9 @@
 /*
  * ncurses-ruby is a ruby module for accessing the FSF's ncurses library
- * (C) 2002 Tobias Peters <t-peters@berlios.de>
- * 
+ * (C) 2002, 2003, 2004 Tobias Peters <t-peters@berlios.de>
+ * (C) 2004 Simon Kaczor <skaczor@cox.net>
+ * (C) 2005 2006 2009 Tobias Herzke
+ *
  *  This module is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
  *  License as published by the Free Software Foundation; either
@@ -16,7 +18,7 @@
  *  License along with this module; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
  *
- * $Id: ncurses_wrap.c,v 1.3 2003/08/28 08:42:24 t-peters Exp $
+ * $Id: ncurses_wrap.c,v 1.16 2009/05/03 20:37:26 t-peters Exp $
  *
  * This file was adapted from the original ncurses header file which
  * has the following copyright statements:
@@ -62,36 +64,15 @@
   - v*printw functions (but normal printw functions are supported!)
 */
 
-#if defined(HAVE_GETWIN) || defined(HAVE_PUTWIN)
-#ifdef HAVE_UNISTD_H
-#include <unistd.h>
-#else
-int dup(int);
-int close(int);
-#endif
+#include "ncurses_wrap.h"
+#ifdef HAVE_LOCALE_H
+#include <locale.h>
 #endif
 
-#ifdef HAVE_NCURSES_H
-#include <ncurses.h>
-#else
-#ifdef HAVE_NCURSES_CURSES_H
-#include <ncurses/curses.h>
-#else
-#include <curses.h>
-#endif
-#endif
-#include <ruby.h>
-
-static VALUE mNcurses;  /* module Ncurses */
-static VALUE cWINDOW;   /* class Ncurses::WINDOW */
-static VALUE cSCREEN;   /* class Ncurses::SCREEN */
-static VALUE eNcurses;  /* Ncurses::Exception thrown by this extension */
-
-#define NCFUNC(name, nargs)                       \
-  rb_define_singleton_method(mNcurses,            \
-                             #name,               \
-                             &rbncurs_ ## name,   \
-                             nargs)
+VALUE mNcurses;  /* module Ncurses */
+VALUE cWINDOW;   /* class Ncurses::WINDOW */
+VALUE cSCREEN;   /* class Ncurses::SCREEN */
+VALUE eNcurses;  /* Ncurses::Exception thrown by this extension */
 
 static void Init_ncurses_full(void);
 
@@ -142,6 +123,17 @@ init_constants_1(void)
     rb_define_const(mNcurses, "WA_TOP",        INT2NUM(WA_TOP));
     rb_define_const(mNcurses, "WA_VERTICAL",   INT2NUM(WA_VERTICAL));
 #endif
+
+    /* locale flags */
+#ifdef HAVE_LOCALE_H
+    rb_define_const(mNcurses, "LC_ALL",        INT2NUM(LC_ALL));
+    rb_define_const(mNcurses, "LC_COLLATE",    INT2NUM(LC_COLLATE));
+    rb_define_const(mNcurses, "LC_CTYPE",      INT2NUM(LC_CTYPE));
+    rb_define_const(mNcurses, "LC_MESSAGES",   INT2NUM(LC_MESSAGES));
+    rb_define_const(mNcurses, "LC_MONETARY",   INT2NUM(LC_MONETARY));
+    rb_define_const(mNcurses, "LC_NUMERIC",    INT2NUM(LC_NUMERIC));
+    rb_define_const(mNcurses, "LC_TIME",       INT2NUM(LC_TIME));
+#endif
 }
 
 static VALUE rbncurs_COLORS()
@@ -215,7 +207,13 @@ init_constants_2(void)
 #endif
 }
 
-static VALUE   wrap_window(WINDOW* window)
+/// Portable (1.8,1.9) determination of array length (calling #length)
+long rbncurs_array_length(VALUE array)
+{
+  return NUM2LONG(funcall(array, rb_intern("length"), 0));
+}
+
+VALUE   wrap_window(WINDOW* window)
 {
     if (window == 0) return Qnil;
     {
@@ -230,7 +228,7 @@ static VALUE   wrap_window(WINDOW* window)
         return rb_window;
     }
 }
-static WINDOW* get_window(VALUE rb_window)
+WINDOW* get_window(VALUE rb_window)
 {
     WINDOW* window;
     if (rb_window == Qnil) return 0;
@@ -341,6 +339,14 @@ static VALUE rbncurs_winnstr(VALUE dummy, VALUE rb_win, VALUE rb_chstr, VALUE rb
 #include "panel_wrap.h" /* needs access to mNcurses, wrap_window, get_window */
 #endif
 
+#ifdef HAVE_FORM_H
+#include "form_wrap.h" /* needs init_form */
+#endif
+
+#ifdef HAVE_MENU_H
+#include "menu_wrap.h" /* needs init_menu */
+#endif
+
 static
 void
 init_functions_0(void)
@@ -398,6 +404,21 @@ static VALUE set_ESCDELAY(VALUE dummy, VALUE new_delay)
     return INT2NUM(ESCDELAY);
 }
 #endif
+
+/* This global is wrapper-specific. It denotes the interval after which the
+   terminal is periodically checked for having resized or not. */
+/* time in milliseconds                           */
+static VALUE get_RESIZEDELAY(){return rb_iv_get(mNcurses, "@resize_delay");}
+static VALUE set_RESIZEDELAY(VALUE dummy, VALUE rb_new_delay)
+{
+    int c_new_delay = NUM2INT(rb_new_delay);
+    if (c_new_delay <= 0)
+        rb_raise(rb_eArgError, "delay must be > 0");
+    rb_new_delay = INT2NUM(c_new_delay);
+    rb_iv_set(mNcurses, "@resize_delay", rb_new_delay);
+    return rb_new_delay ;
+}
+
 static
 void
 init_globals_2(void)
@@ -405,6 +426,7 @@ init_globals_2(void)
     rb_iv_set(mNcurses, "@stdscr", Qnil);
     rb_iv_set(mNcurses, "@curscr", Qnil);
     rb_iv_set(mNcurses, "@newscr", Qnil);
+
     rb_define_module_function(mNcurses, "stdscr", 
                               (&get_stdscr), 0);
     rb_define_module_function(mNcurses, "curscr", 
@@ -427,6 +449,12 @@ init_globals_2(void)
     rb_define_module_function(mNcurses, "ESCDELAY=",
                               (&set_ESCDELAY),1);
 #endif
+    /* The maximum delay before screen resize is detected, in milliseconds */
+    rb_iv_set(mNcurses, "@resize_delay", INT2FIX(333));
+    rb_define_module_function(mNcurses, "RESIZEDELAY",
+                              (&get_RESIZEDELAY),0);
+    rb_define_module_function(mNcurses, "RESIZEDELAY=",
+                              (&set_RESIZEDELAY),1);
 }
 #ifdef HAVE_KEYBOUND
 static VALUE rbncurs_keybound(VALUE dummy, VALUE keycode, VALUE count)
@@ -566,6 +594,7 @@ static VALUE rbncurs_attron(VALUE dummy, VALUE arg1) {
 static VALUE rbncurs_attrset(VALUE dummy, VALUE arg1) {
     return INT2NUM(attrset(NUM2ULONG(arg1)));
 }
+#if defined(NCURSES_VERSION_MAJOR) && NCURSES_VERSION_MAJOR > 4
 #ifdef HAVE_ATTR_OFF
 static VALUE rbncurs_attr_off(VALUE dummy, VALUE arg1, VALUE arg2) {
     return INT2NUM(attr_off(NUM2ULONG(arg1),  ((void)(arg2),NULL)));
@@ -581,6 +610,81 @@ static VALUE rbncurs_attr_set(VALUE dummy, VALUE arg1, VALUE arg2, VALUE arg3) {
     return INT2NUM(attr_set(NUM2ULONG(arg1),  NUM2INT(arg2),  ((void)(arg3),NULL)));
 }
 #endif
+#if defined(HAVE_SLK_ATTR_OFF) || defined(slk_attr_off)
+static VALUE rbncurs_slk_attr_off(VALUE dummy, VALUE arg1, VALUE arg2) {
+    return INT2NUM(slk_attr_off(NUM2ULONG(arg1),  ((void)(arg2),NULL)));
+}
+#endif
+#if defined(HAVE_SLK_ATTR_ON) || defined(slk_attr_on)
+static VALUE rbncurs_slk_attr_on(VALUE dummy, VALUE arg1, VALUE arg2) {
+    return INT2NUM(slk_attr_on(NUM2ULONG(arg1),  ((void)(arg2),NULL)));
+}
+#endif
+#ifdef HAVE_SLK_ATTR_SET
+static VALUE rbncurs_slk_attr_set(VALUE dummy, VALUE arg1, VALUE arg2, VALUE arg3) {
+    return INT2NUM(slk_attr_set(NUM2ULONG(arg1),  NUM2INT(arg2),  ((void)(arg3),NULL)));
+}
+#endif
+#ifdef HAVE_WATTR_ON
+static VALUE rbncurs_wattr_on(VALUE dummy, VALUE arg1, VALUE arg2, VALUE arg3) {
+    return INT2NUM(wattr_on(get_window(arg1),  NUM2ULONG(arg2),  ((void)(arg3),NULL)));
+}
+#endif
+#ifdef HAVE_WATTR_OFF
+static VALUE rbncurs_wattr_off(VALUE dummy, VALUE arg1, VALUE arg2, VALUE arg3) {
+    return INT2NUM(wattr_off(get_window(arg1),  NUM2ULONG(arg2),  ((void)(arg3),NULL)));
+}
+#endif
+#ifdef HAVE_WATTR_SET
+static VALUE rbncurs_wattr_set(VALUE dummy, VALUE arg1, VALUE arg2, VALUE arg3, VALUE arg4) {
+    return INT2NUM(wattr_set(get_window(arg1),  NUM2ULONG(arg2),  NUM2INT(arg3),  ((void)(arg4),NULL)));
+}
+#endif
+#if defined(HAVE_VID_ATTR) || defined(vid_attr)
+static VALUE rbncurs_vid_attr(VALUE dummy, VALUE arg1, VALUE arg2, VALUE arg3) {
+    return INT2NUM(vid_attr(NUM2ULONG(arg1),  NUM2INT(arg2),  ((void)(arg3),NULL)));
+}
+#endif
+#ifdef HAVE_ATTR_GET
+static VALUE rbncurs_attr_get(VALUE dummy, VALUE rb_attrs, VALUE rb_pair,
+                         VALUE dummy2)
+{
+    if ((rb_obj_is_instance_of(rb_attrs, rb_cArray) != Qtrue)
+        || (rb_obj_is_instance_of(rb_pair, rb_cArray) != Qtrue)) {
+        rb_raise(rb_eArgError,
+                 "attrs and pair arguments must be empty Arrays");
+        return Qnil;
+    }
+    {
+        attr_t attrs = 0;
+        short  pair  = 0;
+        int return_value = attr_get(&attrs, &pair, 0);
+        rb_ary_push(rb_attrs, INT2NUM(attrs));
+        rb_ary_push(rb_pair, INT2NUM(pair));
+        return INT2NUM(return_value);
+    }
+}
+static VALUE rbncurs_wattr_get(VALUE dummy,VALUE win, VALUE rb_attrs, VALUE rb_pair,
+                          VALUE dummy2)
+{
+    if ((rb_obj_is_instance_of(rb_attrs, rb_cArray) != Qtrue)
+        || (rb_obj_is_instance_of(rb_pair, rb_cArray) != Qtrue)) {
+        rb_raise(rb_eArgError,
+                 "attrs and pair arguments must be empty Arrays");
+        return Qnil;
+    }
+    {
+        attr_t attrs = 0;
+        short  pair  = 0;
+        int return_value = wattr_get(get_window(win), &attrs, &pair, 0);
+        rb_ary_push(rb_attrs, INT2NUM(attrs));
+        rb_ary_push(rb_pair, INT2NUM(pair));
+        return INT2NUM(return_value);
+    }
+}
+#endif /* HAVE_ATTR_GET */
+#endif
+
 static VALUE rbncurs_baudrate(VALUE dummy) {
     return INT2NUM(baudrate());
 }
@@ -602,8 +706,25 @@ static VALUE rbncurs_box(VALUE dummy, VALUE arg1, VALUE arg2, VALUE arg3) {
 static VALUE rbncurs_can_change_color(VALUE dummy) {
     return (can_change_color()) ? Qtrue : Qfalse;
 }
+
+static int rbncurshelper_halfdelay_cbreak(int tenths, int break_chars)
+{
+    int status = break_chars ? cbreak() : nocbreak();
+    if (status != ERR) {
+        rb_iv_set(mNcurses, "@halfdelay", INT2NUM(tenths));
+        rb_iv_set(mNcurses, "@cbreak", break_chars ? Qtrue : Qfalse);
+    }
+    return status;
+}
+static void rbncurshelper_halfdelay_cbreak_restore()
+{
+    if (RTEST(rb_iv_get(mNcurses, "@cbreak")))
+        cbreak();
+    else
+        nocbreak();
+}
 static VALUE rbncurs_cbreak(VALUE dummy) {
-    return INT2NUM(cbreak());
+    return INT2NUM(rbncurshelper_halfdelay_cbreak(0, 1));
 }
 #ifdef HAVE_CHGAT
 static VALUE rbncurs_chgat(VALUE dummy, VALUE arg1, VALUE arg2, VALUE arg3, VALUE arg4) {
@@ -687,11 +808,58 @@ static VALUE rbncurs_flushinp(VALUE dummy) {
 static VALUE rbncurs_getbkgd(VALUE dummy, VALUE arg1) {
     return INT2NUM(getbkgd(get_window(arg1)));
 }
+
+static int rbncurshelper_nonblocking_wgetch(WINDOW *c_win) {
+    /* nonblocking wgetch only implemented for Ncurses */
+    int halfdelay = NUM2INT(rb_iv_get(mNcurses, "@halfdelay"));
+    int infd = NUM2INT(rb_iv_get(mNcurses, "@infd"));
+    double screen_delay = halfdelay * 0.1;
+#ifdef NCURSES_VERSION
+    int windelay = c_win->_delay;
+#else
+    int windelay = 0;
+#endif
+    double window_delay = (windelay >= 0) ? 0.001 * windelay : (1e200*1e200);
+    /* FIXME:                                                  ^ Infinity ^*/
+    double delay = (screen_delay > 0) ? screen_delay : window_delay;
+    int result;
+    struct timeval tv;
+    struct timezone tz = {0,0};
+    double starttime, nowtime, finishtime;
+    double resize_delay = NUM2INT(get_RESIZEDELAY()) / 1000.0;
+    fd_set in_fds;
+    gettimeofday(&tv, &tz);
+    starttime = tv.tv_sec + tv.tv_usec * 1e-6;
+    finishtime = starttime + delay;
+#ifdef NCURSES_VERSION
+    c_win->_delay = 0;
+#endif
+    while (doupdate() /* detects resize */, (result = wgetch(c_win)) == ERR) {
+        gettimeofday(&tv, &tz);
+        nowtime = tv.tv_sec + tv.tv_usec * 1e-6;
+        delay = finishtime - nowtime;
+	if (delay <= 0) break;
+
+	/* Check for terminal size change every resize_delay seconds */
+        if (resize_delay > delay) resize_delay = delay;
+        tv.tv_sec = (time_t)resize_delay;
+        tv.tv_usec = (unsigned)( (resize_delay - tv.tv_sec) * 1e6 );
+
+	/* sleep on infd until input is available or tv reaches timeout */
+	FD_ZERO(&in_fds);
+	FD_SET(infd, &in_fds);
+	rb_thread_select(infd + 1, &in_fds, NULL, NULL, &tv);
+    }
+#ifdef NCURSES_VERSION
+    c_win->_delay = windelay;
+#endif
+    return result;
+}
 static VALUE rbncurs_getch(VALUE dummy) {
-    return INT2NUM(getch());
+    return INT2NUM(rbncurshelper_nonblocking_wgetch(stdscr));
 }
 static VALUE rbncurs_halfdelay(VALUE dummy, VALUE arg1) {
-    return INT2NUM(halfdelay(NUM2INT(arg1)));
+    return INT2NUM(rbncurshelper_halfdelay_cbreak(NUM2INT(arg1), 1));
 }
 static VALUE rbncurs_has_colors(VALUE dummy) {
     return (has_colors()) ? Qtrue : Qfalse;
@@ -821,6 +989,10 @@ static VALUE rbncurs_initscr(VALUE dummy) {
 #ifdef ACS_SSSS
     rb_define_const(mNcurses, "ACS_SSSS",      INT2NUM(ACS_SSSS));
 #endif
+    rbncurshelper_halfdelay_cbreak_restore();
+    rb_iv_set(mNcurses, "@infd", INT2FIX(0));
+    rb_iv_set(mNcurses, "@halfdelay", INT2FIX(0));
+    rb_iv_set(mNcurses, "@cbreak", Qfalse);
     return v;
 }
 static VALUE rbncurs_init_color(VALUE dummy, VALUE arg1, VALUE arg2, VALUE arg3, VALUE arg4) {
@@ -918,7 +1090,9 @@ static VALUE rbncurs_mvderwin(VALUE dummy, VALUE arg1, VALUE arg2, VALUE arg3) {
     return INT2NUM(mvderwin(get_window(arg1),  NUM2INT(arg2),  NUM2INT(arg3)));
 }
 static VALUE rbncurs_mvgetch(VALUE dummy, VALUE arg1, VALUE arg2) {
-    return INT2NUM(mvgetch(NUM2INT(arg1),  NUM2INT(arg2)));
+    if (wmove(stdscr, NUM2INT(arg1),  NUM2INT(arg2)) == ERR)
+        return INT2NUM(ERR);
+    return INT2NUM(rbncurshelper_nonblocking_wgetch(stdscr));
 }
 #ifdef HAVE_MVHLINE
 static VALUE rbncurs_mvhline(VALUE dummy, VALUE arg1, VALUE arg2, VALUE arg3, VALUE arg4) {
@@ -977,7 +1151,10 @@ static VALUE rbncurs_mvwdelch(VALUE dummy, VALUE arg1, VALUE arg2, VALUE arg3) {
     return INT2NUM(mvwdelch(get_window(arg1),  NUM2INT(arg2),  NUM2INT(arg3)));
 }
 static VALUE rbncurs_mvwgetch(VALUE dummy, VALUE arg1, VALUE arg2, VALUE arg3) {
-    return INT2NUM(mvwgetch(get_window(arg1),  NUM2INT(arg2),  NUM2INT(arg3)));
+    WINDOW * c_win = get_window(arg1);
+    if (wmove(c_win, NUM2INT(arg1),  NUM2INT(arg2)) == ERR)
+        return INT2NUM(ERR);
+    return INT2NUM(rbncurshelper_nonblocking_wgetch(c_win));
 }
 #ifdef HAVE_MVWHLINE
 static VALUE rbncurs_mvwhline(VALUE dummy, VALUE arg1, VALUE arg2, VALUE arg3, VALUE arg4, VALUE arg5) {
@@ -1017,7 +1194,7 @@ static VALUE rbncurs_nl(VALUE dummy) {
     return INT2NUM(nl());
 }
 static VALUE rbncurs_nocbreak(VALUE dummy) {
-    return INT2NUM(nocbreak());
+    return INT2NUM(rbncurshelper_halfdelay_cbreak(0, 0));
 }
 static VALUE rbncurs_nodelay(VALUE dummy, VALUE arg1, VALUE arg2) {
     return INT2NUM(nodelay(get_window(arg1),  RTEST(arg2)));
@@ -1124,25 +1301,24 @@ static VALUE rbncurs_scr_set(VALUE dummy, VALUE arg1) {
 static VALUE rbncurs_setscrreg(VALUE dummy, VALUE arg1, VALUE arg2) {
     return INT2NUM(setscrreg(NUM2INT(arg1),  NUM2INT(arg2)));
 }
-static VALUE rbncurs_set_term(VALUE dummy, VALUE arg1) {
-    return wrap_screen(set_term(get_screen(arg1)));
+static VALUE rbncurs_set_term(VALUE dummy, VALUE rb_new_screen) {
+    VALUE rb_old_screen = wrap_screen(set_term(get_screen(rb_new_screen)));
+    rb_iv_set(rb_old_screen, "@infd",      rb_iv_get(mNcurses, "@infd"));
+    rb_iv_set(rb_old_screen, "@halfdelay", rb_iv_get(mNcurses, "@halfdelay"));
+    rb_iv_set(rb_old_screen, "@cbreak",    rb_iv_get(mNcurses, "@cbreak"));
+    rb_iv_set(mNcurses, "@infd",      rb_iv_get(rb_new_screen, "@infd"));
+    rb_iv_set(mNcurses, "@halfdelay", rb_iv_get(rb_new_screen, "@halfdelay"));
+    rb_iv_set(mNcurses, "@cbreak",    rb_iv_get(rb_new_screen, "@cbreak"));
+
+    rbncurshelper_halfdelay_cbreak_restore();
+    return rb_old_screen;
 }
 static VALUE rbncurs_slk_attroff(VALUE dummy, VALUE arg1) {
     return INT2NUM(slk_attroff(NUM2ULONG(arg1)));
 }
-#if defined(HAVE_SLK_ATTR_OFF) || defined(slk_attr_off)
-static VALUE rbncurs_slk_attr_off(VALUE dummy, VALUE arg1, VALUE arg2) {
-    return INT2NUM(slk_attr_off(NUM2ULONG(arg1),  ((void)(arg2),NULL)));
-}
-#endif
 static VALUE rbncurs_slk_attron(VALUE dummy, VALUE arg1) {
     return INT2NUM(slk_attron(NUM2ULONG(arg1)));
 }
-#if defined(HAVE_SLK_ATTR_ON) || defined(slk_attr_on)
-static VALUE rbncurs_slk_attr_on(VALUE dummy, VALUE arg1, VALUE arg2) {
-    return INT2NUM(slk_attr_on(NUM2ULONG(arg1),  ((void)(arg2),NULL)));
-}
-#endif
 static VALUE rbncurs_slk_attrset(VALUE dummy, VALUE arg1) {
     return INT2NUM(slk_attrset(NUM2ULONG(arg1)));
 }
@@ -1151,11 +1327,7 @@ static VALUE rbncurs_slk_attr(VALUE dummy) {
     return INT2NUM(slk_attr());
 }
 #endif
-#ifdef HAVE_SLK_ATTR_SET
-static VALUE rbncurs_slk_attr_set(VALUE dummy, VALUE arg1, VALUE arg2, VALUE arg3) {
-    return INT2NUM(slk_attr_set(NUM2ULONG(arg1),  NUM2INT(arg2),  ((void)(arg3),NULL)));
-}
-#endif
+
 static VALUE rbncurs_slk_clear(VALUE dummy) {
     return INT2NUM(slk_clear());
 }
@@ -1246,11 +1418,6 @@ static VALUE rbncurs_vidattr(VALUE dummy, VALUE arg1) {
     return INT2NUM(vidattr(NUM2ULONG(arg1)));
 }
 #endif
-#if defined(HAVE_VID_ATTR) || defined(vid_attr)
-static VALUE rbncurs_vid_attr(VALUE dummy, VALUE arg1, VALUE arg2, VALUE arg3) {
-    return INT2NUM(vid_attr(NUM2ULONG(arg1),  NUM2INT(arg2),  ((void)(arg3),NULL)));
-}
-#endif
 static VALUE rbncurs_vline(VALUE dummy, VALUE arg1, VALUE arg2) {
     return INT2NUM(vline(NUM2ULONG(arg1),  NUM2INT(arg2)));
 }
@@ -1285,21 +1452,6 @@ static VALUE rbncurs_wattroff(VALUE dummy, VALUE arg1, VALUE arg2) {
 static VALUE rbncurs_wattrset(VALUE dummy, VALUE arg1, VALUE arg2) {
     return INT2NUM(wattrset(get_window(arg1),  NUM2INT(arg2)));
 }
-#ifdef HAVE_WATTR_ON
-static VALUE rbncurs_wattr_on(VALUE dummy, VALUE arg1, VALUE arg2, VALUE arg3) {
-    return INT2NUM(wattr_on(get_window(arg1),  NUM2ULONG(arg2),  ((void)(arg3),NULL)));
-}
-#endif
-#ifdef HAVE_WATTR_OFF
-static VALUE rbncurs_wattr_off(VALUE dummy, VALUE arg1, VALUE arg2, VALUE arg3) {
-    return INT2NUM(wattr_off(get_window(arg1),  NUM2ULONG(arg2),  ((void)(arg3),NULL)));
-}
-#endif
-#ifdef HAVE_WATTR_SET
-static VALUE rbncurs_wattr_set(VALUE dummy, VALUE arg1, VALUE arg2, VALUE arg3, VALUE arg4) {
-    return INT2NUM(wattr_set(get_window(arg1),  NUM2ULONG(arg2),  NUM2INT(arg3),  ((void)(arg4),NULL)));
-}
-#endif
 static VALUE rbncurs_wbkgd(VALUE dummy, VALUE arg1, VALUE arg2) {
     return INT2NUM(wbkgd(get_window(arg1),  NUM2ULONG(arg2)));
 }
@@ -1344,7 +1496,7 @@ static VALUE rbncurs_werase(VALUE dummy, VALUE arg1) {
     return INT2NUM(werase(get_window(arg1)));
 }
 static VALUE rbncurs_wgetch(VALUE dummy, VALUE arg1) {
-    return INT2NUM(wgetch(get_window(arg1)));
+    return INT2NUM(rbncurshelper_nonblocking_wgetch(get_window(arg1)));
 }
 static VALUE rbncurs_whline(VALUE dummy, VALUE arg1, VALUE arg2, VALUE arg3) {
     return INT2NUM(whline(get_window(arg1),  NUM2ULONG(arg2),  NUM2INT(arg3)));
@@ -1472,8 +1624,16 @@ static VALUE rbncurs_newterm(VALUE dummy, VALUE rb_type, VALUE rb_outfd, VALUE r
     int infd  = NUM2INT(rb_funcall(rb_infd, rb_intern("to_i"), 0));
     VALUE rb_screen =
         wrap_screen(newterm(type, fdopen(outfd, "w"), fdopen(infd, "r")));
-    if (RTEST(rb_screen))
+    if (RTEST(rb_screen)) {
         Init_ncurses_full();
+        rbncurshelper_halfdelay_cbreak_restore();
+    }
+    rb_iv_set(mNcurses, "@infd", INT2NUM(infd));
+    rb_iv_set(rb_screen, "@infd", INT2NUM(infd));
+    rb_iv_set(mNcurses, "@halfdelay", INT2FIX(0));
+    rb_iv_set(rb_screen, "@halfdelay", INT2FIX(0));
+    rb_iv_set(mNcurses, "@cbreak", Qfalse);
+    rb_iv_set(rb_screen, "@cbreak", Qfalse);
     return rb_screen;
 }
 
@@ -1487,6 +1647,7 @@ static void init_functions_2(void) {
     NCFUNC(attroff, 1);
     NCFUNC(attron, 1);
     NCFUNC(attrset, 1);
+#if defined(NCURSES_VERSION_MAJOR) && NCURSES_VERSION_MAJOR > 4
 #ifdef HAVE_ATTR_OFF
     NCFUNC(attr_off, 2);
 #endif
@@ -1495,6 +1656,32 @@ static void init_functions_2(void) {
 #endif
 #ifdef HAVE_ATTR_SET
     NCFUNC(attr_set, 3);
+#endif
+#if defined(HAVE_SLK_ATTR_OFF) || defined(slk_attr_off)
+    NCFUNC(slk_attr_off, 2);
+#endif
+#if defined(HAVE_SLK_ATTR_ON) || defined(slk_attr_on)
+    NCFUNC(slk_attr_on, 2);
+#endif
+#ifdef HAVE_SLK_ATTR_SET
+    NCFUNC(slk_attr_set, 3);
+#endif
+#ifdef HAVE_WATTR_ON
+   NCFUNC(wattr_on, 3);
+#endif
+#ifdef HAVE_WATTR_OFF
+    NCFUNC(wattr_off, 3);
+#endif
+#ifdef HAVE_WATTR_SET
+    NCFUNC(wattr_set, 4);
+#endif
+#if defined (HAVE_VID_ATTR) || defined(vid_attr)
+    NCFUNC(vid_attr, 3);
+#endif
+#ifdef HAVE_ATTR_GET
+    NCFUNC(attr_get, 3);
+    NCFUNC(wattr_get, 4);
+#endif /* HAVE_ATTR_GET */
 #endif
     NCFUNC(baudrate, 0);
     NCFUNC(beep, 0);
@@ -1673,19 +1860,10 @@ static void init_functions_2(void) {
     NCFUNC(setscrreg, 2);
     NCFUNC(set_term, 1);
     NCFUNC(slk_attroff, 1);
-#if defined(HAVE_SLK_ATTR_OFF) || defined(slk_attr_off)
-    NCFUNC(slk_attr_off, 2);
-#endif
     NCFUNC(slk_attron, 1);
-#if defined(HAVE_SLK_ATTR_ON) || defined(slk_attr_on)
-    NCFUNC(slk_attr_on, 2);
-#endif
     NCFUNC(slk_attrset, 1);
 #ifdef HAVE_SLK_ATTR
     NCFUNC(slk_attr, 0);
-#endif
-#ifdef HAVE_SLK_ATTR_SET
-    NCFUNC(slk_attr_set, 3);
 #endif
     NCFUNC(slk_clear, 0);
 #ifdef HAVE_SLK_COLOR
@@ -1720,9 +1898,6 @@ static void init_functions_2(void) {
     NCFUNC(untouchwin, 1);
 #ifdef HAVE_VIDATTR
     NCFUNC(vidattr, 1);
-#endif
-#if defined (HAVE_VID_ATTR) || defined(vid_attr)
-    NCFUNC(vid_attr, 3);
 #endif
     NCFUNC(vline, 2);
     NCFUNC(waddch, 2);
@@ -1733,15 +1908,6 @@ static void init_functions_2(void) {
     NCFUNC(wattron, 2);
     NCFUNC(wattroff, 2);
     NCFUNC(wattrset, 2);
- #ifdef HAVE_WATTR_ON
-   NCFUNC(wattr_on, 3);
-#endif
-#ifdef HAVE_WATTR_OFF
-    NCFUNC(wattr_off, 3);
-#endif
-#ifdef HAVE_WATTR_SET
-    NCFUNC(wattr_set, 4);
-#endif
     NCFUNC(wbkgd, 2);
     NCFUNC(wbkgdset, 2);
     NCFUNC(wborder, 9);
@@ -1837,7 +2003,7 @@ static void init_constants_3(void) {
     /* Pseudo-character tokens outside ASCII range.  The curses wgetch()
      * function will return any given one of these only if the corresponding
      * k- capability is defined in your terminal's terminfo entry. */
-#ifdef KAY_CODE_YES
+#ifdef KEY_CODE_YES
     rb_define_const(mNcurses, "KEY_CODE_YES", INT2NUM(KEY_CODE_YES));
 #endif
     rb_define_const(mNcurses, "KEY_MIN", INT2NUM(KEY_MIN));
@@ -2220,8 +2386,9 @@ static VALUE rbncurs_wprintw(int argc, VALUE * argv, VALUE dummy)
 
 /* Debugging : use with libncurses_g.a */
 #ifdef HAVE__TRACEF
-static VALUE rbncurs_tracef(int argc, VALUE * argv, VALUE)
+static VALUE rbncurs_tracef(int argc, VALUE * argv, VALUE dummy)
 {
+    (void) dummy;
     if (argc < 1) {
         rb_raise(rb_eArgError, "function needs at least 1 argument");
         return Qnil;
@@ -2232,7 +2399,7 @@ static VALUE rbncurs_tracef(int argc, VALUE * argv, VALUE)
 }
 #endif /* HAVE__TRACEF */
 #ifdef HAVE__TRACEDUMP
-static VALUE rbncurs_tracedump(VALUE dummy, VALUE rb_label, label rb_win)
+static VALUE rbncurs_tracedump(VALUE dummy, VALUE rb_label, VALUE rb_win)
 {
     _tracedump(STR2CSTR(rb_label), get_window(rb_win));
 }
@@ -2279,51 +2446,13 @@ static VALUE rbncurs_trace(VALUE dummy, VALUE param)
 #endif /* HAVE_TRACE */
 #ifdef HAVE__NC_TRACEBITS
 static VALUE rbncurs_nc_tracebits()
-{ return rb_str_new2(_nc_tracebits()); }
+{ return rb_str_new2((char*)_nc_tracebits()); }
 #endif /* HAVE__NC_TRACEBITS */
 
 #ifdef HAVE_ASSUME_DEFAULT_COLORS
 static VALUE rbncurs_assume_default_colors(VALUE dummy, VALUE fg, VALUE bg)
 { return INT2NUM(assume_default_colors(NUM2INT(fg),NUM2INT(bg))); }
 #endif  /* HAVE_ASSUME_DEFAULT_COLORS */
-#ifdef HAVE_ATTR_GET
-static VALUE rbncurs_attr_get(VALUE dummy, VALUE rb_attrs, VALUE rb_pair,
-                         VALUE dummy2)
-{
-    if ((rb_obj_is_instance_of(rb_attrs, rb_cArray) != Qtrue)
-        || (rb_obj_is_instance_of(rb_pair, rb_cArray) != Qtrue)) {
-        rb_raise(rb_eArgError,
-                 "attrs and pair arguments must be empty Arrays");
-        return Qnil;
-    }
-    {
-        attr_t attrs = 0;
-        short  pair  = 0;
-        int return_value = attr_get(&attrs, &pair, 0);
-        rb_ary_push(rb_attrs, INT2NUM(attrs));
-        rb_ary_push(rb_pair, INT2NUM(pair));
-        return INT2NUM(return_value);
-    }
-}
-static VALUE rbncurs_wattr_get(VALUE dummy,VALUE win, VALUE rb_attrs, VALUE rb_pair,
-                          VALUE dummy2)
-{
-    if ((rb_obj_is_instance_of(rb_attrs, rb_cArray) != Qtrue)
-        || (rb_obj_is_instance_of(rb_pair, rb_cArray) != Qtrue)) {
-        rb_raise(rb_eArgError,
-                 "attrs and pair arguments must be empty Arrays");
-        return Qnil;
-    }
-    {
-        attr_t attrs = 0;
-        short  pair  = 0;
-        int return_value = wattr_get(get_window(win), &attrs, &pair, 0);
-        rb_ary_push(rb_attrs, INT2NUM(attrs));
-        rb_ary_push(rb_pair, INT2NUM(pair));
-        return INT2NUM(return_value);
-    }
-}
-#endif /* HAVE_ATTR_GET */
 
 static void init_functions_3(void)
 {
@@ -2332,7 +2461,7 @@ static void init_functions_3(void)
     NCFUNC(ungetmouse, 1);
 #endif
 #ifdef HAVE_MOUSEMASK
-    NCFUNC(mousemask, 1);
+    NCFUNC(mousemask, 2);
 #endif
 #ifdef HAVE_WENCLOSE
     rb_define_module_function(mNcurses, "wenclose?",
@@ -2414,10 +2543,6 @@ static void init_functions_3(void)
 #ifdef HAVE_ASSUME_DEFAULT_COLORS
     NCFUNC(assume_default_colors, 2);
 #endif  /* HAVE_ASSUME_DEFAULT_COLORS */
-#ifdef HAVE_ATTR_GET
-    NCFUNC(attr_get, 3);
-    NCFUNC(wattr_get, 4);
-#endif /* HAVE_ATTR_GET */
     NCFUNC(wprintw, -1);
 }
 
@@ -2472,10 +2597,10 @@ static void init_constants_4(void)
 #define rb_ACS(ACS)                               \
 VALUE rb_## ACS (VALUE rb_screen)                 \
 {                                                 \
-    SCREEN * c_screen = get_screen(rb_screen);    \
-    SCREEN * current_screen = set_term(c_screen); \
+    VALUE current_screen =                        \
+      rbncurs_set_term(mNcurses, rb_screen);      \
     VALUE rb_ACS_CONST = INT2NUM(ACS);            \
-    set_term(current_screen);                     \
+    rbncurs_set_term(mNcurses, current_screen);   \
     return  rb_ACS_CONST;                         \
 }             
 #define wrap_ACS(ACS)                                          \
@@ -2579,6 +2704,11 @@ void init_SCREEN_methods(void)
 #endif
 }
 
+#ifdef HAVE_LOCALE_H
+static VALUE rbncurs_setlocale(VALUE dummy, VALUE category, VALUE locale)
+{   return rb_str_new2(setlocale(NUM2INT(category), STR2CSTR(locale)));}
+#endif
+
 static void init_safe_functions(void)
 {
     NCFUNC(initscr, 0);
@@ -2590,13 +2720,24 @@ static void init_safe_functions(void)
 #ifdef HAVE_USE_ENV
     NCFUNC(use_env, 1);
 #endif
+#ifdef HAVE_LOCALE_H
+    NCFUNC(setlocale, 2);
+#endif
 }
-void Init_ncurses(void)
+void Init_ncurses_bin(void)
 {
     mNcurses = rb_define_module("Ncurses");
     eNcurses = rb_define_class_under(mNcurses, "Exception", rb_eRuntimeError);
     rb_iv_set(mNcurses, "@windows_hash", rb_hash_new());
     rb_iv_set(mNcurses, "@screens_hash", rb_hash_new());
+
+    /* keep track of "halfdelay" settings in the wrapper */
+    rb_iv_set(mNcurses, "@halfdelay", INT2FIX(0));
+    rb_iv_set(mNcurses, "@cbreak", Qfalse);
+
+    /* filedescriptor that transports input from terminal to application */
+    rb_iv_set(mNcurses, "@infd", Qnil);
+
     cWINDOW  = rb_define_class_under(mNcurses, "WINDOW", rb_cObject);
     cSCREEN  = rb_define_class_under(mNcurses, "SCREEN", rb_cObject);
     init_constants_1();
@@ -2618,4 +2759,10 @@ static void Init_ncurses_full(void)
 #ifdef HAVE_PANEL_H
     init_panel();
 #endif
+#ifdef HAVE_FORM_H
+    init_form();
+#endif
+#ifdef HAVE_MENU_H
+    init_menu();
+#endif
 }
